import { isGitIgnoredSync } from 'globby';
import micromatch from 'micromatch';
import { _getDependenciesFromScripts } from './binaries/index.js';
import { ConfigurationChief } from './ConfigurationChief.js';
import { ConsoleStreamer } from './ConsoleStreamer.js';
import { DependencyDeputy } from './DependencyDeputy.js';
import { IssueCollector } from './IssueCollector.js';
import { IssueFixer } from './IssueFixer.js';
import { PrincipalFactory } from './PrincipalFactory.js';
import { ProjectPrincipal } from './ProjectPrincipal.js';
import { debugLogObject, debugLogArray, debugLog } from './util/debug.js';
import { _glob, negate } from './util/glob.js';
import { getEntryPathFromManifest, getPackageNameFromFilePath, getPackageNameFromModuleSpecifier, normalizeSpecifierFromFilePath, } from './util/modules.js';
import { dirname, isInNodeModules, join, isInternal } from './util/path.js';
import { fromBinary, isBinary } from './util/protocols.js';
import { _resolveSpecifier } from './util/require.js';
import { loadTSConfig } from './util/tsconfig-loader.js';
import { WorkspaceWorker } from './WorkspaceWorker.js';
export const main = async (unresolvedConfiguration) => {
    const { cwd, tsConfigFile, gitignore, isStrict, isProduction, isShowProgress, isIncludeEntryExports, isIsolateWorkspaces, isFix, fixTypes, } = unresolvedConfiguration;
    debugLogObject('*', 'Unresolved configuration (from CLI arguments)', unresolvedConfiguration);
    const chief = new ConfigurationChief({ cwd, isProduction, isStrict, isIncludeEntryExports });
    const deputy = new DependencyDeputy({ isProduction, isStrict });
    const factory = new PrincipalFactory();
    const streamer = new ConsoleStreamer({ isEnabled: isShowProgress });
    const isGitIgnored = gitignore ? isGitIgnoredSync({ cwd }) : () => false;
    streamer.cast('Reading workspace configuration(s)...');
    await chief.init();
    const compilers = chief.getCompilers();
    const workspaces = chief.getWorkspaces();
    const report = chief.getIssueTypesToReport();
    const rules = chief.getRules();
    const filters = chief.getFilters();
    const fixer = new IssueFixer({ isEnabled: isFix, cwd, fixTypes });
    const isReportDependencies = report.dependencies || report.unlisted || report.unresolved;
    const isReportValues = report.exports || report.nsExports || report.classMembers;
    const isReportTypes = report.types || report.nsTypes || report.enumMembers;
    const collector = new IssueCollector({ cwd, rules, filters });
    const enabledPluginsStore = new Map();
    deputy.addIgnored(chief.config.ignoreBinaries, chief.config.ignoreDependencies);
    const o = () => workspaces.map(w => ({ pkgName: w.pkgName, name: w.name, config: w.config, ancestors: w.ancestors }));
    debugLogObject('*', 'Included workspaces', () => workspaces.map(w => w.pkgName));
    debugLogObject('*', 'Included workspace configs', o);
    const handleReferencedDependency = ({ specifier, containingFilePath, principal, workspace, }) => {
        if (isInternal(specifier)) {
            const filePath = principal.resolveModule(specifier, containingFilePath)?.resolvedFileName;
            if (filePath) {
                const ignorePatterns = workspace.config.ignore.map(pattern => join(workspace.dir, pattern));
                const isIgnored = micromatch.isMatch(filePath, ignorePatterns);
                if (!isIgnored)
                    principal.addEntryPath(filePath);
            }
            else {
                collector.addIssue({ type: 'unresolved', filePath: containingFilePath, symbol: specifier });
            }
        }
        else {
            if (isBinary(specifier)) {
                const binaryName = fromBinary(specifier);
                const isHandled = deputy.maybeAddReferencedBinary(workspace, binaryName);
                if (!isHandled)
                    collector.addIssue({ type: 'binaries', filePath: containingFilePath, symbol: binaryName });
            }
            else {
                const packageName = isInNodeModules(specifier)
                    ? getPackageNameFromFilePath(specifier)
                    : getPackageNameFromModuleSpecifier(specifier);
                const isHandled = packageName && deputy.maybeAddReferencedExternalDependency(workspace, packageName);
                if (!isHandled)
                    collector.addIssue({ type: 'unlisted', filePath: containingFilePath, symbol: specifier });
                if (packageName && specifier !== packageName) {
                    const otherWorkspace = chief.availableWorkspaceManifests.find(w => w.manifest.name === packageName);
                    if (otherWorkspace) {
                        const filePath = _resolveSpecifier(otherWorkspace.dir, normalizeSpecifierFromFilePath(specifier));
                        if (filePath) {
                            principal.addEntryPath(filePath, { skipExportsAnalysis: true });
                        }
                        else {
                            collector.addIssue({ type: 'unresolved', filePath: containingFilePath, symbol: specifier });
                        }
                    }
                }
            }
        }
    };
    for (const workspace of workspaces) {
        const { name, dir, config, ancestors, pkgName, manifestPath, manifest } = workspace;
        const { paths, ignoreDependencies, ignoreBinaries } = config;
        streamer.cast(`Analyzing workspace ${name}...`);
        deputy.addWorkspace({ name, dir, manifestPath, manifest, ignoreDependencies, ignoreBinaries });
        const { compilerOptions, definitionPaths } = await loadTSConfig(join(dir, tsConfigFile ?? 'tsconfig.json'));
        const principal = factory.getPrincipal({
            cwd: dir,
            paths,
            compilerOptions,
            compilers,
            pkgName,
            isGitIgnored,
            isIsolateWorkspaces,
        });
        const worker = new WorkspaceWorker({
            name,
            dir,
            cwd,
            config,
            manifest,
            isProduction,
            isStrict,
            rootIgnore: chief.config.ignore,
            negatedWorkspacePatterns: chief.getNegatedWorkspacePatterns(name),
            enabledPluginsInAncestors: ancestors.flatMap(ancestor => enabledPluginsStore.get(ancestor) ?? []),
        });
        await worker.init();
        principal.addEntryPaths(definitionPaths);
        debugLogArray(name, `Definition paths`, definitionPaths);
        const sharedGlobOptions = { cwd, workingDir: dir, gitignore, ignore: worker.getIgnorePatterns() };
        const entryPathsFromManifest = await getEntryPathFromManifest(manifest, sharedGlobOptions);
        debugLogArray(name, 'Entry paths in package.json', entryPathsFromManifest);
        principal.addEntryPaths(entryPathsFromManifest);
        const dependencies = await worker.findAllDependencies();
        const { referencedDependencies, hostDependencies, installedBinaries, hasTypesIncluded, enabledPlugins, entryFilePatterns, productionEntryFilePatterns, } = dependencies;
        deputy.addHostDependencies(name, hostDependencies);
        deputy.setInstalledBinaries(name, installedBinaries);
        deputy.setHasTypesIncluded(name, hasTypesIncluded);
        enabledPluginsStore.set(name, enabledPlugins);
        referencedDependencies.forEach(([containingFilePath, specifier]) => {
            handleReferencedDependency({ specifier, containingFilePath, principal, workspace });
        });
        if (isProduction) {
            const negatedEntryPatterns = entryFilePatterns.map(negate);
            {
                const patterns = worker.getProductionEntryFilePatterns(negatedEntryPatterns);
                const workspaceEntryPaths = await _glob({ ...sharedGlobOptions, patterns });
                debugLogArray(name, `Entry paths`, workspaceEntryPaths);
                principal.addEntryPaths(workspaceEntryPaths);
            }
            {
                const pluginWorkspaceEntryPaths = await _glob({ ...sharedGlobOptions, patterns: productionEntryFilePatterns });
                debugLogArray(name, `Production plugin entry paths`, pluginWorkspaceEntryPaths);
                principal.addEntryPaths(pluginWorkspaceEntryPaths, { skipExportsAnalysis: true });
            }
            {
                const patterns = worker.getProductionProjectFilePatterns(negatedEntryPatterns);
                const workspaceProjectPaths = await _glob({ ...sharedGlobOptions, patterns });
                debugLogArray(name, `Project paths`, workspaceProjectPaths);
                workspaceProjectPaths.forEach(projectPath => principal.addProjectPath(projectPath));
            }
        }
        else {
            {
                const patterns = worker.getEntryFilePatterns();
                const workspaceEntryPaths = await _glob({ ...sharedGlobOptions, patterns });
                debugLogArray(name, `Entry paths`, workspaceEntryPaths);
                principal.addEntryPaths(workspaceEntryPaths);
            }
            {
                const patterns = worker.getProjectFilePatterns([...productionEntryFilePatterns]);
                const workspaceProjectPaths = await _glob({ ...sharedGlobOptions, patterns });
                debugLogArray(name, `Project paths`, workspaceProjectPaths);
                workspaceProjectPaths.forEach(projectPath => principal.addProjectPath(projectPath));
            }
            {
                const patterns = [...entryFilePatterns, ...productionEntryFilePatterns];
                const pluginWorkspaceEntryPaths = await _glob({ ...sharedGlobOptions, patterns });
                debugLogArray(name, `Plugin entry paths`, pluginWorkspaceEntryPaths);
                principal.addEntryPaths(pluginWorkspaceEntryPaths, { skipExportsAnalysis: true });
            }
            {
                const patterns = worker.getPluginProjectFilePatterns();
                const pluginWorkspaceProjectPaths = await _glob({ ...sharedGlobOptions, patterns });
                debugLogArray(name, `Plugin project paths`, pluginWorkspaceProjectPaths);
                pluginWorkspaceProjectPaths.forEach(projectPath => principal.addProjectPath(projectPath));
            }
            {
                const patterns = worker.getPluginConfigPatterns();
                const configurationEntryPaths = await _glob({ ...sharedGlobOptions, patterns });
                debugLogArray(name, `Plugin configuration paths`, configurationEntryPaths);
                principal.addEntryPaths(configurationEntryPaths, { skipExportsAnalysis: true });
            }
        }
        if (chief.resolvedConfigFilePath) {
            principal.addEntryPath(chief.resolvedConfigFilePath, { skipExportsAnalysis: true });
        }
    }
    const principals = factory.getPrincipals();
    debugLog('*', `Created ${principals.length} programs for ${workspaces.length} workspaces`);
    const analyzedFiles = new Set();
    const exportedSymbols = {};
    const importedSymbols = {};
    const unreferencedFiles = new Set();
    const entryPaths = new Set();
    for (const principal of principals) {
        const specifierFilePaths = new Set();
        const analyzeSourceFile = (filePath, _principal = principal) => {
            const workspace = chief.findWorkspaceByFilePath(filePath);
            if (workspace) {
                const { imports, exports, scripts } = _principal.analyzeSourceFile(filePath, {
                    skipTypeOnly: isStrict,
                    isFixExports: fixer.isFixUnusedExports,
                    isFixTypes: fixer.isFixUnusedTypes,
                    ignoreExportsUsedInFile: Boolean(chief.config.ignoreExportsUsedInFile),
                });
                const { internal, external, unresolved } = imports;
                const { exported, duplicate } = exports;
                if (Object.keys(exported).length > 0)
                    exportedSymbols[filePath] = exported;
                for (const [specifierFilePath, importItems] of Object.entries(internal)) {
                    const packageName = getPackageNameFromModuleSpecifier(importItems.specifier);
                    if (packageName && chief.availableWorkspacePkgNames.has(packageName)) {
                        external.add(packageName);
                        if (_principal === principal) {
                            const workspace = chief.findWorkspaceByFilePath(specifierFilePath);
                            if (workspace) {
                                const principal = factory.getPrincipalByPackageName(workspace.pkgName);
                                if (principal && !principal.isGitIgnored(specifierFilePath)) {
                                    specifierFilePaths.add(specifierFilePath);
                                }
                            }
                        }
                    }
                    if (!importedSymbols[specifierFilePath]) {
                        importedSymbols[specifierFilePath] = importItems;
                    }
                    else {
                        const importedModule = importedSymbols[specifierFilePath];
                        for (const id of importItems.symbols)
                            importedModule.symbols.push(id);
                        for (const id of importItems.importedNs)
                            importedModule.importedNs.add(id);
                        for (const id of importItems.isReExportedBy)
                            importedModule.isReExportedBy.add(id);
                        for (const id of importItems.isImportedBy)
                            importedModule.isImportedBy.add(id);
                        if (importItems.hasStar)
                            importedModule.hasStar = true;
                        if (importItems.isReExport) {
                            importedModule.isReExport = true;
                            importedModule.isReExportedBy.add(filePath);
                        }
                    }
                }
                duplicate.forEach(symbols => {
                    if (symbols.length > 1) {
                        const symbol = symbols.map(s => s.symbol).join('|');
                        collector.addIssue({ type: 'duplicates', filePath, symbol, symbols });
                    }
                });
                external.forEach(specifier => {
                    const packageName = getPackageNameFromModuleSpecifier(specifier);
                    const isHandled = packageName && deputy.maybeAddReferencedExternalDependency(workspace, packageName);
                    if (!isHandled)
                        collector.addIssue({ type: 'unlisted', filePath, symbol: specifier });
                });
                unresolved.forEach(unresolvedImport => {
                    const { specifier, pos, line, col } = unresolvedImport;
                    collector.addIssue({ type: 'unresolved', filePath, symbol: specifier, pos, line, col });
                });
                _getDependenciesFromScripts(scripts, { cwd: dirname(filePath) }).forEach(specifier => {
                    handleReferencedDependency({ specifier, containingFilePath: filePath, principal: _principal, workspace });
                });
            }
        };
        streamer.cast('Running async compilers...');
        await principal.runAsyncCompilers();
        streamer.cast('Connecting the dots...');
        let size = principal.entryPaths.size;
        let round = 0;
        do {
            size = principal.entryPaths.size;
            const resolvedFiles = principal.getUsedResolvedFiles();
            const files = resolvedFiles.filter(filePath => !analyzedFiles.has(filePath));
            debugLogArray('*', `Analyzing used resolved files [P${principals.indexOf(principal) + 1}/${++round}]`, files);
            files.forEach(filePath => {
                analyzeSourceFile(filePath);
                analyzedFiles.add(filePath);
            });
        } while (size !== principal.entryPaths.size);
        specifierFilePaths.forEach(specifierFilePath => {
            if (!analyzedFiles.has(specifierFilePath)) {
                analyzedFiles.add(specifierFilePath);
                analyzeSourceFile(specifierFilePath, principal);
            }
        });
        principal.getUnreferencedFiles().forEach(filePath => unreferencedFiles.add(filePath));
        principal.entryPaths.forEach(filePath => entryPaths.add(filePath));
        factory.deletePrincipal(principal);
    }
    const isSymbolImported = (filePath, symbol, importsForExport, depth = 0) => {
        if (!importsForExport)
            return false;
        if (importsForExport.symbols.includes(symbol))
            return true;
        if (Array.from(importsForExport.importedNs).find(ns => importsForExport.symbols.includes(`${ns}.${symbol}`)))
            return true;
        if (Array.from(importsForExport.isReExportedAs).find(([file, ns]) => isSymbolImported(filePath, `${ns}.${symbol}`, importedSymbols[file], depth + 1)))
            return true;
        const { isReExport, isReExportedBy } = importsForExport;
        const hasSymbol = (file) => isSymbolImported(filePath, symbol, importedSymbols[file], depth + 1);
        if (isReExport)
            return Array.from(isReExportedBy).some(hasSymbol);
        return false;
    };
    const isExportedInEntryFile = (importedModule) => {
        if (!importedModule)
            return false;
        const { isReExport, isReExportedBy } = importedModule;
        const hasFile = (file) => entryPaths.has(file) || isExportedInEntryFile(importedSymbols[file]);
        return isReExport ? Array.from(isReExportedBy).some(hasFile) : false;
    };
    const getReExportingEntryFile = (importedModule, symbol) => {
        if (!importedModule)
            return undefined;
        const { isReExport, isReExportedBy } = importedModule;
        if (isReExport) {
            for (const f of isReExportedBy) {
                if (entryPaths.has(f) && f in exportedSymbols && symbol in exportedSymbols[f])
                    return f;
                else
                    return getReExportingEntryFile(importedSymbols[f], symbol);
            }
        }
    };
    const isExportedItemReferenced = (exportedItem) => {
        return (exportedItem.refs > 0 &&
            (typeof chief.config.ignoreExportsUsedInFile === 'object'
                ? exportedItem.type !== 'unknown' && !!chief.config.ignoreExportsUsedInFile[exportedItem.type]
                : chief.config.ignoreExportsUsedInFile));
    };
    if (isReportValues || isReportTypes) {
        streamer.cast('Analyzing source files...');
        for (const [filePath, exportItems] of Object.entries(exportedSymbols)) {
            const workspace = chief.findWorkspaceByFilePath(filePath);
            if (workspace) {
                const { isIncludeEntryExports } = workspace.config;
                if (!isIncludeEntryExports && entryPaths.has(filePath))
                    continue;
                const importsForExport = importedSymbols[filePath];
                for (const [symbol, exportedItem] of Object.entries(exportItems)) {
                    if (exportedItem.jsDocTags.includes('@public') || exportedItem.jsDocTags.includes('@beta'))
                        continue;
                    if (exportedItem.jsDocTags.includes('@alias'))
                        continue;
                    if (isProduction && exportedItem.jsDocTags.includes('@internal'))
                        continue;
                    if (importsForExport && isSymbolImported(filePath, symbol, importsForExport)) {
                        if (!isIncludeEntryExports && importsForExport.isReExport && isExportedInEntryFile(importsForExport)) {
                            continue;
                        }
                        if (exportedItem.type === 'enum') {
                            exportedItem.members?.forEach(member => {
                                if (member.refs === 0 &&
                                    !isSymbolImported(filePath, `${symbol}.${member.identifier}`, importsForExport)) {
                                    collector.addIssue({
                                        type: 'enumMembers',
                                        filePath,
                                        symbol: member.identifier,
                                        parentSymbol: symbol,
                                        pos: exportedItem.pos,
                                        line: exportedItem.line,
                                        col: exportedItem.col,
                                    });
                                }
                            });
                        }
                        continue;
                    }
                    const hasNsImport = Boolean(importsForExport?.hasStar);
                    const reExportingEntryFile = getReExportingEntryFile(importsForExport, symbol);
                    if (reExportingEntryFile)
                        continue;
                    if (!isIncludeEntryExports && importsForExport && importsForExport.hasStar && importsForExport.isReExport) {
                        continue;
                    }
                    const isType = ['enum', 'type', 'interface'].includes(exportedItem.type);
                    if (hasNsImport && ((!report.nsTypes && isType) || (!report.nsExports && !isType)))
                        continue;
                    if (!isExportedItemReferenced(exportedItem)) {
                        const type = isType ? (hasNsImport ? 'nsTypes' : 'types') : hasNsImport ? 'nsExports' : 'exports';
                        collector.addIssue({
                            type,
                            filePath,
                            symbol,
                            symbolType: exportedItem.type,
                            pos: exportedItem.pos,
                            line: exportedItem.line,
                            col: exportedItem.col,
                        });
                        if (isType)
                            fixer.addUnusedTypeNode(filePath, exportedItem.fixes);
                        else
                            fixer.addUnusedExportNode(filePath, exportedItem.fixes);
                    }
                }
            }
        }
    }
    const unusedFiles = [...unreferencedFiles].filter(filePath => !analyzedFiles.has(filePath));
    collector.addFilesIssues(unusedFiles);
    collector.addFileCounts({ processed: analyzedFiles.size, unused: unusedFiles.length });
    if (isReportDependencies) {
        const { dependencyIssues, devDependencyIssues, optionalPeerDependencyIssues } = deputy.settleDependencyIssues();
        const { configurationHints } = deputy.getConfigurationHints();
        dependencyIssues.forEach(issue => collector.addIssue(issue));
        if (!isProduction)
            devDependencyIssues.forEach(issue => collector.addIssue(issue));
        optionalPeerDependencyIssues.forEach(issue => collector.addIssue(issue));
        configurationHints.forEach(hint => collector.addConfigurationHint(hint));
    }
    const unusedIgnoredWorkspaces = chief.getUnusedIgnoredWorkspaces();
    unusedIgnoredWorkspaces.forEach(identifier => collector.addConfigurationHint({ type: 'ignoreWorkspaces', identifier }));
    const { issues, counters, configurationHints } = collector.getIssues();
    if (isFix) {
        await fixer.fixIssues(issues);
    }
    streamer.clear();
    return { report, issues, counters, rules, configurationHints };
};
